
/*
svcmgr.exe - command line application for controlling Windows NT services. 
Version 0.1
User can list active and/or inactive services and driver services, display dependent 
service information and start/stop services on the local/remote machines.

This code can be compiled with GNU gcc 2.95.2 mingw32 version and Borland C++ 5.5 free compiler. 
Not tested on Microsoft Visual C++.

Copyright (C) 2000  Antti Markus (antti@pld.ttu.ee)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <iostream>
using std::cout;
using std::cerr;
using std::string;

#include <windows.h>
#include "svcmgr.h"

const DWORD TYPE_ALL = SERVICE_WIN32 | SERVICE_DRIVER;

const int ACTION_LIST = 1;
const int ACTION_SHOWINFO = 2;
const int ACTION_DEPENDENT = 4;
const int ACTION_STOP = 8;
const int ACTION_START = 16; 

const int RESULT_OK = 0;
const int RESULT_ACCESS_DENIED = 1;
const int RESULT_DB_OPEN_FAILED = 2;
const int RESULT_INVALID_OPTION = 3;
const int RESULT_OTHER_ERROR = 4;
const int RESULT_OPERATION_FAILED = 5;


int main(int argc, char *argv[])
{   
  string strMachineName; 
  string strServiceName;          
  string strTmp;             
  DWORD dwSvcState=0;
  DWORD dwSvcType=0;
  int iAction=0;
  
  //extract command line arguments from command line
  for(int i=0; i < argc; i++)
  {  
    //char *pTmp=argv[i];
    if(argv[i][0] == '-')
    {
      //argument is option
      strTmp=argv[i];
      if(strTmp == "-service")
      {                            
        //list only WIN32 services
        dwSvcType=SERVICE_WIN32;
      }                         
      if(strTmp == "-driver")
      {                         
        if(dwSvcType != 0)
        {                         
          //we can list both driver and WIN32 services
          dwSvcType|=SERVICE_DRIVER;
        }                           
        else
        {                       
          //list only driver services
          dwSvcType=SERVICE_DRIVER;
        }
      }                                 
      if(strTmp == "-both")
      {                              
        //list driver and WIN32 services
        dwSvcType=TYPE_ALL;
      } 
      if(strTmp == "-active")     
      {                                 
        //list active services
        dwSvcState=SERVICE_ACTIVE;
      }
      if(strTmp == "-inactive")
      {
        if(dwSvcState != 0)
        {                         
          //list inactive and active services
          dwSvcState|=SERVICE_INACTIVE;
        }
        else
        {                     
          //list inactive services         
          dwSvcState=SERVICE_INACTIVE;
        }
      }                               
      if(strTmp == "-all")
      {                                      
        //list inactive and active services
        dwSvcState=SERVICE_STATE_ALL;
      }                        
      
      if(strTmp == "-list")
      {                                    
        //list services
        iAction=ACTION_LIST;
      }                       
      if((strTmp == "-dependent") ||
         (strTmp == "-dep"))
      {
        if(iAction != 0)
        {              
          //list services and their dependent services
          iAction|=ACTION_DEPENDENT;
        }                           
        else
        {
          iAction=ACTION_DEPENDENT;
        }
      }                   
      
      if(strTmp == "-stop")
      {                                
        if(iAction & ACTION_START)
        {
          cerr << "Incompatible options: -stop, -start" << endl;
          return RESULT_INVALID_OPTION;
        }                                             
        //stop sepcified service
        iAction=ACTION_STOP;
      } 
      if(strTmp == "-start")
      {                   
        if(iAction & ACTION_STOP)
        {
          cerr << "Incompatible options: -stop, -start" << endl;
          return RESULT_INVALID_OPTION;
        }
        //start specified service
        iAction=ACTION_START;
      }   

    }                     
    else if(argv[i][0] == '\\')
    {
      //argument is machine name
      strMachineName=argv[i];
    }                   
    else
    {   
      //argument is service name
      if(i != 0)
      {
        strServiceName=argv[i];
      }  
    }
  }                                                         
  
  //set default values          
  //(if they're not already set through command line)
  if(dwSvcType == 0)
  {
    dwSvcType=SERVICE_WIN32;
  }                                  
  
  if(dwSvcState == 0)
  {
    dwSvcState=SERVICE_ACTIVE;
  }                      
  
  if((iAction == 0) || (iAction == ACTION_DEPENDENT))
  {
    iAction|=ACTION_LIST;
  }
  
  //open service control manager                 
  TServiceControlManager TestManager;
  if(TestManager.Open(GENERIC_READ, NULL, ((strMachineName == "")?NULL:strMachineName.c_str())) == FALSE)
  {            
    if(TestManager.GetErrorCode() == ERROR_ACCESS_DENIED)
    {                                                            
      cerr << "Error: insufficient user rights" << endl;    
      cerr << "Error: access to service control manager denied" << endl;
      return RESULT_ACCESS_DENIED;
    }                                                  
    else
    {
      cerr << "Error: could not open service control manager" << endl;
      return RESULT_DB_OPEN_FAILED;
    }  
  }
  else
  {                
    
    if(iAction & ACTION_LIST)
    {
      if(TestManager.EnumerateStatus(dwSvcState, dwSvcType) == TRUE)
      {
        //change output according to options
        string strStr;
        switch(dwSvcState)
        {
          case SERVICE_ACTIVE:
            strStr=" active";
            break;
            
          case SERVICE_INACTIVE:
            strStr=" inactive";
            break;
            
          case SERVICE_STATE_ALL:
            strStr="";
            break;    
            
        } 
        
        switch(dwSvcType)
        {
          case SERVICE_DRIVER:
            strStr+=" driver service";
            break;
            
          case SERVICE_WIN32:
          case TYPE_ALL:
            strStr+=" service";
            break;  
        }               
        
        if(TestManager.GetServiceCount() > 0)
        {           
          //plural if there is more than one service
          strStr+="s";
        }
                       
        cout << TestManager.GetServiceCount() << strStr << " on " << ((TestManager.GetMachine()=="")?static_cast<string>("localhost"):TestManager.GetMachine()) << endl; 
        TServiceStatus StatusInfo;
        
        //repeat while there are services...
        while(TestManager.GetNext(StatusInfo) != FALSE)
        {                                   
          //display service name and service display (user friendly) name
          cout << StatusInfo.lpServiceName << " (" << StatusInfo.lpDisplayName << ") " << ((dwSvcState == SERVICE_STATE_ALL)?StatusInfo.GetState():static_cast<string>("")) << endl;
          if(iAction & ACTION_DEPENDENT)
          {
            DWORD dwCount=0;                                                             
            ENUM_SERVICE_STATUS *pBuf=static_cast<ENUM_SERVICE_STATUS*>(NULL);
            pBuf=TestManager.GetDependentServices(StatusInfo.lpServiceName, dwCount);

            if(pBuf != NULL)
            {
              cout << "\t" << dwCount << " dependent services:" << endl;
              for(DWORD i=0; i < dwCount; i++)
              {
                cout << "\t" << pBuf[i].lpServiceName << " (" << pBuf[i].lpDisplayName << ") " << static_cast<TServiceStatus>(&pBuf[i]).GetState() << endl;
              } 
              cout << endl;
              
              delete [] pBuf;
              pBuf=static_cast<TServiceStatus*>(NULL);
            } //end if(pBuf != NULL)
          } //end if(iAction & ACTION_DEPENDENT) 
        } //end while(TestManager.GetNext(...  
      }
      else                             
      {                                
        cerr << "Error: enumerating services failed" << endl;
        return RESULT_OTHER_ERROR;
      }
    } //end if(iAction & ACTION_LIST)
    else if(iAction & ACTION_STOP)
    {                                                   
      if(strServiceName == "")
      {
        cerr << "Error: service not specified" << endl;
        return RESULT_INVALID_OPTION;
      }
      
      cout << "Stopping " << strServiceName << " service on " << ((strMachineName == "")?"localhost":strMachineName.c_str())  << "..." << endl;
      BOOL bResult=TestManager.Stop(strServiceName.c_str());
      if(bResult == TRUE)
      {
        cout << "Stopped" << endl;
      }
      else
      {          
        cout << "Error: service not stopped" << endl;
        
        if(TestManager.GetErrorCode() == ERROR_DEPENDENT_SERVICES_RUNNING)
        {                                
          cerr << "Error: dependent services still running" << endl;
          return RESULT_OPERATION_FAILED;
        }
        else if(TestManager.GetErrorCode() == ERROR_ACCESS_DENIED)
        {  
          cerr << "Error: access denied - insufficient user rights" << endl;
          return RESULT_ACCESS_DENIED;
        }                                                         
        else
        {                                
          return RESULT_OPERATION_FAILED;
        }
      }
    } //end else if(iAction & ACTION_STOP) 
    else if(iAction & ACTION_START)
    {                         
      if(strServiceName == "")
      {
        cerr << "Error: service not specified" << endl;
        return RESULT_INVALID_OPTION;
      }
      
      cout << "Starting " << strServiceName << " service on " << ((strMachineName == "")?"localhost":strMachineName.c_str()) << "..." << endl;
      BOOL bResult=TestManager.Start(strServiceName.c_str());
      if(bResult == TRUE)
      {
        cout << "Started" << endl;
      }
      else
      {          
        cout << "Error: service not started" << endl;                     
        if(TestManager.GetErrorCode() != ERROR_ACCESS_DENIED)
        {
          return RESULT_OPERATION_FAILED;
        }
        else
        {  
          cerr << "Error: access denied - insufficient user rights" << endl;
          return RESULT_ACCESS_DENIED;
        }        
      }
    } //end else if(iAction & ACTION_START)
  }  //end if(TestManager.Open(GENERIC_READ, NULL, ... else ...     
  
  return RESULT_OK;
}